# -*- coding: utf-8 -*-

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals
)


import copy
import six


class PytingoData(object):
    """Settings flexible mapper"""

    __slots__ = [
        "_data_dict", "_meta_field", "_separator", "_wildcard", "_list_accessor", "_meta", "_processed_data_dict"
    ]

    def __init__(self, data_dict, process_meta=True, meta_field=".meta", separator=".", wildcard="*", list_accessor="$"):
        """
        Args:
            :param data_dict:
            :type data_dict:

            :param process_meta:
            :type process_meta:

            :param meta_field:
            :type meta_field:

            :param separator:
            :type separator:

            :param wildcard:
            :type wildcard:
        """
        if not isinstance(data_dict, dict):
            raise ValueError("data_dict must be of type dict.")
        if not isinstance(process_meta, bool):
            raise ValueError("process_meta must be of type bool.")

        if not isinstance(meta_field, six.string_types):
            raise ValueError("meta_field must be of type str.")
        if not isinstance(separator, six.string_types):
            raise ValueError("separator must be of type str.")
        if not isinstance(wildcard, six.string_types):
            raise ValueError("wildcard must be of type str.")
        if not isinstance(list_accessor, six.string_types):
            raise ValueError("list_accessor must be of type str.")

        # We make a deep copy to avoid side effects if data_dict changes.
        self._data_dict = copy.deepcopy(data_dict)

        self._meta_field = meta_field
        self._separator = separator
        self._wildcard = wildcard
        self._list_accessor = list_accessor

        if not process_meta or self._meta_field not in self._data_dict:
            self._meta = {}
            self._processed_data_dict = self._data_dict
        else:
            self._meta = self._data_dict.pop(self._meta_field)
            self._process_meta()

    def _process_meta(self):
        """
        """
        # We make a deep copy to modify self._processed_data_dict without any risks.
        self._processed_data_dict = copy.deepcopy(self._data_dict)

        if "inheritance" in self._meta:
            self._process_inheritance()

    def _process_inheritance(self):
        """
        """
        inheritance = self._meta["inheritance"]

        for inheritor_str_path in inheritance:
            inheritors = self._get_inheritors(
                inheritor_str_path.split(self._separator)
            )

            for inheritor in inheritors:
                if isinstance(inheritor, dict):
                    for inheritable_value_key in inheritance[inheritor_str_path]:
                        if inheritable_value_key in inheritor:
                            continue

                        try:
                            inheritor[inheritable_value_key] = self.get(inheritable_value_key)
                        except KeyError:
                            pass
                elif isinstance(inheritor, list):
                    pass
                else:
                    # TODO: Raise Custom Exception
                    raise ValueError(
                        "A field pointed by the path '" + inheritor_str_path +
                        "' is an atomic value and can't inherit anything."
                    )

    def _get_inheritors(self, path, tracepoint=None):
        if tracepoint is None:
            if path == []:
                raise RuntimeError("There is no point to make . inheriting from a subfield.")
            tracepoint = self._processed_data_dict

        key = path[0]

        if isinstance(tracepoint, dict):
            if key == self._wildcard:
                inheritors = []
                selected_items_type = None

                for tp_key in tracepoint:
                    item_type = type(tracepoint[tp_key])

                    if selected_items_type is None:
                        selected_items_type = item_type
                    elif selected_items_type != item_type:
                        raise ValueError(
                            "The items types must be uniform when the items are selected with a wildcard."
                        )

                    if len(path) == 1:
                        inheritors.append(tracepoint[tp_key])
                    else:
                        inheritors.extend(self._get_inheritors(
                            path[1:], tracepoint[tp_key]
                        ))

                return inheritors

            elif key in tracepoint:
                if len(path) == 1:
                    return [tracepoint[key]]

                return self._get_inheritors(path[1:], tracepoint[key])
            else:
                return []

        elif isinstance(tracepoint, list):
            inheritors = []
            selected_items_type = None

            for item in tracepoint:
                item_type = type(item)

                if selected_items_type is None:
                    selected_items_type = item_type
                elif selected_items_type != item_type:
                    raise ValueError(
                        "The items types must be uniform when the items are selected with a wildcard."
                    )

                inheritors.extend(self._get_inheritors(
                    (path if key != self._wildcard else path[1:]), item
                ))

            return inheritors

        else:
            return []

    def as_dict(self):
        # We make a copy to avoid trolling from outside the object.
        return copy.deepcopy(self._processed_data_dict)

    def __iter__(self):
        # That's useful to iterate over dict-like settings fields.
        return self._processed_data_dict.__iter__()

    def get(self, key):
        """
        This method gives us a value of our SettingsData object.
        If the given key refers to an intermediate subdictionary, then
        a new SettingsData object is returned.
        """

        value = self._processed_data_dict
        for _key in key.split(self._separator):
            if _key != '':
                if isinstance(value, dict):
                    value = value[_key]
                elif isinstance(value, list):
                    value = value[int(_key)]

        if isinstance(value, dict):
            return PytingoData(
                value, False, self._meta_field, self._separator, self._wildcard, self._list_accessor
            )
        else:
            return value

    def __getattr__(self, attr):
        """
        Convenient accessor.
        """
        return self.get(attr)

    def __getitem__(self, item):
        """
        Convenient accessor.
        """
        return self.get(item)

    def __eq__(self, b):
        for k in self._processed_data_dict:
            if self.get(k) != b.get(k):
                return False

        return True

    def __ne__(self, b):
        return not self.__eq__(b)

    def __contains__(self, item):
        try:
            self.get(item)

            return True
        except KeyError:
            return False
